# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
import torch
from torch import nn
from torch.nn.functional import gelu

from nemo.collections.common.parts import form_attention_mask

__all__ = ["TransformerEmbedding", "AttentionBridge"]


class FixedPositionalEncoding(nn.Module):
    """
    Fixed positional encoding (embedding layer) from sine and cosine functions
    of different frequencies according to https://arxiv.org/abs/1706.03762

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        max_sequence_length: maximum allowed length of the input sequence
    """

    def __init__(self, hidden_size, max_sequence_length=512):
        super().__init__()

        self._hidden_size = hidden_size
        self._max_sequence_length = max_sequence_length
        self._build_pos_enc(hidden_size=self._hidden_size, max_sequence_length=self._max_sequence_length)

    def _build_pos_enc(self, hidden_size, max_sequence_length, device=None):
        """
        Builds/replaces pre-computed positional encoding.
        """
        pos_enc = torch.zeros(max_sequence_length, hidden_size, device=device)
        position = torch.arange(0.0, max_sequence_length).unsqueeze(1)
        coef = -math.log(10000.0) / hidden_size
        div_term = torch.exp(coef * torch.arange(0.0, hidden_size, 2))
        pos_enc[:, 0::2] = torch.sin(position * div_term)
        pos_enc[:, 1::2] = torch.cos(position * div_term)
        pos_enc.div_(math.sqrt(hidden_size))
        self.register_buffer('pos_enc', pos_enc)

    def forward(self, position_ids):
        embeddings = torch.embedding(self.pos_enc, position_ids)
        return embeddings


class TransformerEmbedding(nn.Module):
    """
    Embedding from token and position embeddings.
    Optionally add token_type embedding (e.g. type of the sentence in BERT).

    Args:
        vocab_size: size of the vocabulary
        hidden_size: size of the embeddings in the model, also known as d_model
        max_sequence_length: maximum allowed length of the input sequence
        num_token_types: number of different token types
            (e.g. tokens of sentence A and tokens of sentence B in BERT)
        embedding_dropout: probability of dropout applied to embeddings
        learn_positional_encodings: whether to learn positional encodings or
            use fixed (sine-cosine) ones
    """

    def __init__(
        self,
        vocab_size,
        hidden_size,
        max_sequence_length=512,
        num_token_types=2,
        embedding_dropout=0.0,
        learn_positional_encodings=False,
    ):
        super().__init__()

        self.max_sequence_length = max_sequence_length
        self.learn_positional_encodings = learn_positional_encodings
        self.token_embedding = nn.Embedding(vocab_size, hidden_size, padding_idx=0)
        if learn_positional_encodings:
            self.position_embedding = nn.Embedding(max_sequence_length, hidden_size)
        else:
            self.position_embedding = FixedPositionalEncoding(hidden_size, max_sequence_length)
        if num_token_types > 0:
            self.token_type_embedding = nn.Embedding(num_token_types, hidden_size)
        self.layer_norm = nn.LayerNorm(hidden_size, eps=1e-5)
        self.dropout = nn.Dropout(embedding_dropout)

    def forward(self, input_ids, token_type_ids=None, start_pos=0):
        seq_length = input_ids.size(1)
        # we fail here only with parametric positional embedding. FixedPositionalEncoding automatically extends.
        if self.learn_positional_encodings and (seq_length > self.max_sequence_length):
            raise ValueError(
                f"Input sequence is longer than maximum allowed sequence length for positional encoding. "
                f"Got {seq_length} and {self.max_sequence_length}"
            )

        # prepare position embedding for asynchronius decoding (canary streaming)
        if torch.is_tensor(start_pos):
            shift_pos = start_pos.unsqueeze(-1)
            start_pos = 0
        else:
            shift_pos = None

        position_ids = torch.arange(
            start=start_pos, end=start_pos + seq_length, dtype=torch.long, device=input_ids.device
        )
        position_ids = position_ids.unsqueeze(0).repeat(input_ids.size(0), 1)
        if torch.is_tensor(shift_pos):
            # shift_pos is a tensor, so we need to add it to the position_ids
            # and make sure that the resulting position_ids are within the
            # range of the positional embedding
            position_ids = position_ids + shift_pos

        token_embeddings = self.token_embedding(input_ids)
        position_embeddings = self.position_embedding(position_ids)
        embeddings = token_embeddings + position_embeddings

        if token_type_ids is not None:
            token_type_embeddings = self.token_type_embedding(token_type_ids)
            embeddings = embeddings + token_type_embeddings

        embeddings = self.layer_norm(embeddings)
        embeddings = self.dropout(embeddings)

        return embeddings


class MultiHeadAttention(nn.Module):
    """
    Multi-head scaled dot-product attention layer.

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        num_attention_heads: number of heads in multi-head attention
        attn_score_dropout: probability of dropout applied to attention scores
        attn_layer_dropout: probability of dropout applied to the output of the
            whole layer, but before layer normalization
    """

    def __init__(
        self,
        hidden_size,
        num_attention_heads,
        attn_score_dropout=0.0,
        attn_layer_dropout=0.0,
        return_xatt_scores=False,
    ):
        super().__init__()
        if hidden_size % num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number "
                "of attention heads (%d)" % (hidden_size, num_attention_heads)
            )
        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads
        self.attn_head_size = int(hidden_size / num_attention_heads)
        self.attn_scale = math.sqrt(math.sqrt(self.attn_head_size))

        self.query_net = nn.Linear(hidden_size, hidden_size)
        self.key_net = nn.Linear(hidden_size, hidden_size)
        self.value_net = nn.Linear(hidden_size, hidden_size)
        self.out_projection = nn.Linear(hidden_size, hidden_size)

        self.attn_dropout = nn.Dropout(attn_score_dropout)
        self.layer_dropout = nn.Dropout(attn_layer_dropout)
        self.return_xatt_scores = return_xatt_scores

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attn_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, queries, keys, values, attention_mask):

        # attention_mask is needed to hide the tokens which correspond to [PAD]
        # in the case of BERT, or to hide the future tokens in the case of
        # vanilla language modeling and translation
        query = self.query_net(queries)
        key = self.key_net(keys)
        value = self.value_net(values)
        query = self.transpose_for_scores(query) / self.attn_scale
        key = self.transpose_for_scores(key) / self.attn_scale
        value = self.transpose_for_scores(value)

        # for numerical stability we pre-divide query and key by sqrt(sqrt(d))
        attention_scores = torch.matmul(query, key.transpose(-1, -2))
        if attention_mask is not None:
            attention_scores = attention_scores + attention_mask.to(attention_scores.dtype)
        attention_probs = torch.softmax(attention_scores, dim=-1)
        attention_probs = self.attn_dropout(attention_probs)

        context = torch.matmul(attention_probs, value)
        context_hidden_size = context.size()[-1] * self.num_attention_heads
        context = context.permute(0, 2, 1, 3).contiguous()
        new_context_shape = context.size()[:-2] + (context_hidden_size,)
        context = context.view(*new_context_shape)

        # output projection
        output_states = self.out_projection(context)
        output_states = self.layer_dropout(output_states)

        extra_output = {}
        if self.return_xatt_scores:
            extra_output['xatt_scores'] = attention_probs

        return output_states, extra_output


class PositionWiseFF(nn.Module):
    """
    Position-wise feed-forward network of Transformer block.

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        inner_size: number of neurons in the intermediate part of feed-forward
            net, usually is (4-8 x hidden_size) in the papers
        ffn_dropout: probability of dropout applied to net output
        hidden_act: activation function used between two linear layers
    """

    def __init__(self, hidden_size, inner_size, ffn_dropout=0.0, hidden_act="relu"):
        super().__init__()
        self.dense_in = nn.Linear(hidden_size, inner_size)
        self.dense_out = nn.Linear(inner_size, hidden_size)
        self.layer_dropout = nn.Dropout(ffn_dropout)
        ACT2FN = {"gelu": gelu, "relu": torch.relu}
        self.act_fn = ACT2FN[hidden_act]

    def forward(self, hidden_states):
        output_states = self.dense_in(hidden_states)
        output_states = self.act_fn(output_states)
        output_states = self.dense_out(output_states)
        output_states = self.layer_dropout(output_states)
        return output_states


class AttentionBridge(torch.nn.Module):
    """
    A multi-head attention bridge to project a variable-size hidden states
    to k hidden states (per attention head).

    Code is based on the paper https://arxiv.org/pdf/1703.03130.pdf
    """

    def __init__(self, hidden_size, k, bridge_size):
        """
        hidden_size - size of input hidden state
        k - number of attention heads
        bridge_size - size of internal feed forward weights (i.e., attention head size)
        """
        super().__init__()

        self.hidden_size = hidden_size
        self.k = k
        self.bridge_size = bridge_size

        self.attn_scale = np.sqrt(np.sqrt(self.bridge_size))

        # build model

        self.W1 = torch.nn.Linear(hidden_size, bridge_size, bias=False)
        self.W2 = torch.nn.Linear(bridge_size, k, bias=False)
        self.act = torch.nn.ReLU()

    def forward(self, hidden, hidden_mask=None, return_ortho_loss=False):
        """
        Project hidden [B x N x H] to fixed-size [B x k x H]

        return_ortho_loss - if True returns loss term to encourage
                              orthogonal attention vectors
        """

        attention_scores = self.W2(self.act(self.W1(hidden) / self.attn_scale) / self.attn_scale).transpose(-1, -2)

        attention_mask = form_attention_mask(hidden_mask)
        if attention_mask is not None:
            attention_mask.squeeze_(1)
            attention_scores = attention_scores + attention_mask.to(attention_scores.dtype)

        A = torch.softmax(attention_scores, dim=-1)
        M = A @ hidden

        if return_ortho_loss:
            ortho_loss = ((A @ A.transpose(-1, -2)) - torch.eye(self.k).type_as(A)).pow(2).sum()

            return M, ortho_loss
        else:
            return M



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING

import torch

from nemo.collections.llm.gpt.model.gemma3 import (
    Gemma3Config,
    Gemma3Config1B,
    Gemma3Config4B,
    Gemma3Config12B,
    Gemma3Config27B,
)
from nemo.collections.vlm.gemma3vl.model.base import Gemma3VLConfig, Gemma3VLModel
from nemo.collections.vlm.gemma3vl.model.vision import Gemma3VLMultimodalProjectorConfig, Gemma3VLVisionConfig
from nemo.collections.vlm.neva.model.llava import export_qkv, export_qkv_bias, import_qkv
from nemo.lightning import io, teardown
from nemo.lightning.io.state import TransformFns

if TYPE_CHECKING:
    from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer


@dataclass
class Gemma3VLConfig4B(Gemma3VLConfig):
    """Gemma3 VL config 4B"""

    language_transformer_config: Gemma3Config = field(default_factory=lambda: Gemma3Config4B())
    vision_transformer_config: Gemma3VLVisionConfig = field(default_factory=lambda: Gemma3VLVisionConfig())
    vision_projection_config: Gemma3VLMultimodalProjectorConfig = field(
        default_factory=lambda: Gemma3VLMultimodalProjectorConfig(input_size=1152, hidden_size=2560)
    )


@dataclass
class Gemma3VLConfig12B(Gemma3VLConfig):
    """Gemma3 VL config 12B"""

    language_transformer_config: Gemma3Config = field(default_factory=lambda: Gemma3Config12B())
    vision_transformer_config: Gemma3VLVisionConfig = field(default_factory=lambda: Gemma3VLVisionConfig())
    vision_projection_config: Gemma3VLMultimodalProjectorConfig = field(
        default_factory=lambda: Gemma3VLMultimodalProjectorConfig(input_size=1152, hidden_size=3840)
    )


@dataclass
class Gemma3VLConfig27B(Gemma3VLConfig):
    """Gemma3 VL config 27B"""

    language_transformer_config: Gemma3Config = field(default_factory=lambda: Gemma3Config27B())
    vision_transformer_config: Gemma3VLVisionConfig = field(default_factory=lambda: Gemma3VLVisionConfig())
    vision_projection_config: Gemma3VLMultimodalProjectorConfig = field(
        default_factory=lambda: Gemma3VLMultimodalProjectorConfig(input_size=1152, hidden_size=5376)
    )


@io.model_importer(Gemma3VLModel, "hf")
class Gemma3VLImporter(io.ModelConnector["Gemma3Model", Gemma3VLModel]):
    """Gemma3 VL model HF importer"""

    def init(self) -> Gemma3VLModel:
        return Gemma3VLModel(self.config, tokenizer=self.tokenizer)

    def apply(self, output_path: Path) -> Path:
        # pylint: disable=C0115,C0116
        from transformers import Gemma3Model

        source = Gemma3Model.from_pretrained(str(self))
        target = self.init()
        trainer = self.nemo_setup(target)
        self.convert_state(source, target)
        self.nemo_save(output_path, trainer)

        print(f"Converted HF Gemma3VL model to NeMo, saved to {output_path}")

        teardown(trainer, target)
        del trainer, target

        return output_path

    def convert_state(self, source, target):
        # pylint: disable=C0301,C0116
        mapping = {
            # vision model
            "vision_tower.vision_model.embeddings.patch_embedding.weight": "vision_model.conv1.weight",
            "vision_tower.vision_model.embeddings.patch_embedding.bias": "vision_model.conv1.bias",
            "vision_tower.vision_model.embeddings.position_embedding.weight": "vision_model.position_embeddings.weight",
            "vision_tower.vision_model.post_layernorm.weight": "vision_model.ln_post.weight",
            "vision_tower.vision_model.post_layernorm.bias": "vision_model.ln_post.bias",
            "vision_tower.vision_model.encoder.layers.*.layer_norm1.weight": "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight",
            "vision_tower.vision_model.encoder.layers.*.layer_norm1.bias": "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_bias",
            "vision_tower.vision_model.encoder.layers.*.self_attn.out_proj.weight": "vision_model.decoder.layers.*.self_attention.linear_proj.weight",
            "vision_tower.vision_model.encoder.layers.*.self_attn.out_proj.bias": "vision_model.decoder.layers.*.self_attention.linear_proj.bias",
            "vision_tower.vision_model.encoder.layers.*.layer_norm2.weight": "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight",
            "vision_tower.vision_model.encoder.layers.*.layer_norm2.bias": "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_bias",
            "vision_tower.vision_model.encoder.layers.*.mlp.fc1.weight": "vision_model.decoder.layers.*.mlp.linear_fc1.weight",
            "vision_tower.vision_model.encoder.layers.*.mlp.fc1.bias": "vision_model.decoder.layers.*.mlp.linear_fc1.bias",
            "vision_tower.vision_model.encoder.layers.*.mlp.fc2.weight": "vision_model.decoder.layers.*.mlp.linear_fc2.weight",
            "vision_tower.vision_model.encoder.layers.*.mlp.fc2.bias": "vision_model.decoder.layers.*.mlp.linear_fc2.bias",
            # vision projector
            "multi_modal_projector.mm_soft_emb_norm.weight": "vision_projection.mm_soft_embed_norm.weight",
            # text model
            "language_model.embed_tokens.weight": "language_model.embedding.word_embeddings.weight",
            "language_model.layers.*.input_layernorm.weight": "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight",
            "language_model.layers.*.self_attn.q_norm.weight": "language_model.decoder.layers.*.self_attention.q_layernorm.weight",
            "language_model.layers.*.self_attn.k_norm.weight": "language_model.decoder.layers.*.self_attention.k_layernorm.weight",
            "language_model.layers.*.self_attn.o_proj.weight": "language_model.decoder.layers.*.self_attention.linear_proj.weight",
            "language_model.layers.*.post_attention_layernorm.weight": (
                "language_model.decoder.layers.*.self_attention.linear_proj.post_layernorm.weight"
            ),
            "language_model.layers.*.pre_feedforward_layernorm.weight": "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight",
            "language_model.layers.*.mlp.down_proj.weight": "language_model.decoder.layers.*.mlp.linear_fc2.weight",
            "language_model.layers.*.post_feedforward_layernorm.weight": (
                "language_model.decoder.layers.*.mlp.linear_fc2.post_layernorm.weight"
            ),
            "language_model.norm.weight": "language_model.decoder.final_layernorm.weight",
        }
        transforms = [
            _import_vision_qkv,
            _import_vision_qkv_bias,
            _import_language_qkv,
            io.state_transform(
                source_key="multi_modal_projector.mm_input_projection_weight",
                target_key="vision_projection.proj.weight",
                fn=_vision_projector_permute,
            ),
            io.state_transform(
                source_key=(
                    "language_model.layers.*.mlp.gate_proj.weight",
                    "language_model.layers.*.mlp.up_proj.weight",
                ),
                target_key="language_model.decoder.layers.*.mlp.linear_fc1.weight",
                fn=TransformFns.merge_fc1,
            ),
        ]
        return io.apply_transforms(source, target, mapping=mapping, transforms=transforms)

    @property
    def tokenizer(self) -> "AutoTokenizer":
        # pylint: disable=C0115,C0116
        from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer

        return AutoTokenizer(str(self))

    @property
    def config(self) -> Gemma3VLConfig:
        # pylint: disable=C0115,C0116
        from transformers import Gemma3Config as HFGemma3Config

        name = str(self)
        source = HFGemma3Config.from_pretrained(name)
        source_text = source.text_config
        source_vision = source.vision_config

        if source_text.num_hidden_layers == 26:
            language_transformer_config = Gemma3Config1B()
        elif source_text.num_hidden_layers == 34:
            language_transformer_config = Gemma3Config4B()
        elif source_text.num_hidden_layers == 48:
            language_transformer_config = Gemma3Config12B()
        elif source_text.num_hidden_layers == 62:
            language_transformer_config = Gemma3Config27B()
        else:
            raise ValueError(f"Unrecognized import model: {name}")
        vision_transformer_config = Gemma3VLVisionConfig()
        vision_projection_config = Gemma3VLMultimodalProjectorConfig(
            input_size=source_vision.hidden_size,
            hidden_size=source_text.hidden_size,
        )

        output = Gemma3VLConfig(
            language_transformer_config=language_transformer_config,
            vision_transformer_config=vision_transformer_config,
            vision_projection_config=vision_projection_config,
        )
        return output


@io.model_exporter(Gemma3VLModel, "hf")
class Gemma3VLExporter(io.ModelConnector[Gemma3VLModel, "Gemma3Model"]):
    """Export Gemma3 VL to HF"""

    def init(self):
        from transformers import Gemma3Model
        from transformers.modeling_utils import no_init_weights

        with no_init_weights():
            return Gemma3Model._from_config(self.config)

    def apply(self, output_path: Path) -> Path:
        # pylint: disable=C0115,C0116
        target = self.init()
        source, _ = self.nemo_load(str(self))
        target = self.convert_state(source, target)

        target = target.cpu()
        target.save_pretrained(output_path)
        self.tokenizer.save_pretrained(output_path)

    def convert_state(self, source, target):
        # pylint: disable=C0115,C0116,C0301
        mapping = {
            # vision model
            "vision_model.conv1.weight": "vision_tower.vision_model.embeddings.patch_embedding.weight",
            "vision_model.conv1.bias": "vision_tower.vision_model.embeddings.patch_embedding.bias",
            "vision_model.position_embeddings.weight": "vision_tower.vision_model.embeddings.position_embedding.weight",
            "vision_model.ln_post.weight": "vision_tower.vision_model.post_layernorm.weight",
            "vision_model.ln_post.bias": "vision_tower.vision_model.post_layernorm.bias",
            "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "vision_tower.vision_model.encoder.layers.*.layer_norm1.weight",
            "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_bias": "vision_tower.vision_model.encoder.layers.*.layer_norm1.bias",
            "vision_model.decoder.layers.*.self_attention.linear_proj.weight": "vision_tower.vision_model.encoder.layers.*.self_attn.out_proj.weight",
            "vision_model.decoder.layers.*.self_attention.linear_proj.bias": "vision_tower.vision_model.encoder.layers.*.self_attn.out_proj.bias",
            "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "vision_tower.vision_model.encoder.layers.*.layer_norm2.weight",
            "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_bias": "vision_tower.vision_model.encoder.layers.*.layer_norm2.bias",
            "vision_model.decoder.layers.*.mlp.linear_fc1.weight": "vision_tower.vision_model.encoder.layers.*.mlp.fc1.weight",
            "vision_model.decoder.layers.*.mlp.linear_fc1.bias": "vision_tower.vision_model.encoder.layers.*.mlp.fc1.bias",
            "vision_model.decoder.layers.*.mlp.linear_fc2.weight": "vision_tower.vision_model.encoder.layers.*.mlp.fc2.weight",
            "vision_model.decoder.layers.*.mlp.linear_fc2.bias": "vision_tower.vision_model.encoder.layers.*.mlp.fc2.bias",
            # vision projector
            "vision_projection.mm_soft_embed_norm.weight": "multi_modal_projector.mm_soft_emb_norm.weight",
            # text model
            "language_model.embedding.word_embeddings.weight": "language_model.embed_tokens.weight",
            "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "language_model.layers.*.input_layernorm.weight",
            "language_model.decoder.layers.*.self_attention.q_layernorm.weight": "language_model.layers.*.self_attn.q_norm.weight",
            "language_model.decoder.layers.*.self_attention.k_layernorm.weight": "language_model.layers.*.self_attn.k_norm.weight",
            "language_model.decoder.layers.*.self_attention.linear_proj.weight": "language_model.layers.*.self_attn.o_proj.weight",
            "language_model.decoder.layers.*.self_attention.linear_proj.post_layernorm.weight": (
                "language_model.layers.*.post_attention_layernorm.weight"
            ),
            "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "language_model.layers.*.pre_feedforward_layernorm.weight",
            "language_model.decoder.layers.*.mlp.linear_fc2.weight": "language_model.layers.*.mlp.down_proj.weight",
            "language_model.decoder.layers.*.mlp.linear_fc2.post_layernorm.weight": (
                "language_model.layers.*.post_feedforward_layernorm.weight"
            ),
            "language_model.decoder.final_layernorm.weight": "language_model.norm.weight",
        }

        transforms = [
            _export_vision_qkv,
            _export_vision_qkv_bias,
            _export_language_qkv,
            io.state_transform(
                source_key="vision_projection.proj.weight",
                target_key="multi_modal_projector.mm_input_projection_weight",
                fn=_vision_projector_permute,
            ),
            io.state_transform(
                source_key="language_model.decoder.layers.*.mlp.linear_fc1.weight",
                target_key=(
                    "language_model.layers.*.mlp.gate_proj.weight",
                    "language_model.layers.*.mlp.up_proj.weight",
                ),
                fn=TransformFns.split_fc1,
            ),
        ]
        return io.apply_transforms(source, target, mapping=mapping, transforms=transforms)

    @property
    def tokenizer(self):
        # pylint: disable=C0115,C0116
        return io.load_context(str(self)).model.tokenizer.tokenizer

    @property
    def config(self):
        # pylint: disable=C0115,C0116
        source: Gemma3VLConfig = io.load_context(str(self), subpath="model.config")
        source_text: Gemma3Config = source.language_transformer_config
        source_vision: Gemma3VLVisionConfig = source.vision_transformer_config

        from transformers import Gemma3Config as HFGemma3Config
        from transformers import Gemma3TextConfig as HFGemma3TextConfig
        from transformers import SiglipVisionConfig as HFGemma3VisionConfig

        output_text = HFGemma3TextConfig(
            architectures=["Gemma3ForCausalLM"],
            num_hidden_layers=source_text.num_layers,
            hidden_size=source_text.hidden_size,
            intermediate_size=source_text.ffn_hidden_size,
            num_attention_heads=source_text.num_attention_heads,
            head_dim=source_text.kv_channels,
            hidden_activation="gelu_pytorch_tanh",
            max_position_embeddings=source_text.seq_length,
            initializer_range=source_text.init_method_std,
            rms_norm_eps=source_text.layernorm_epsilon,
            num_key_value_heads=source_text.num_query_groups,
            vocab_size=source_text.vocab_size,
            rope_theta=source_text.rotary_base[1],
            rope_local_base_freq=source_text.rotary_base[0],
        )
        if source_text.num_layers == 62:  # 27B
            output_text.query_pre_attn_scalar = 168
        else:
            output_text.query_pre_attn_scalar = output_text.head_dim

        output_vision = HFGemma3VisionConfig(
            hidden_size=source_vision.hidden_size,
            image_size=source_vision.img_h,
            intermediate_size=source_vision.ffn_hidden_size,
            num_attention_heads=source_vision.num_attention_heads,
            num_hidden_layers=source_vision.num_layers,
            patch_size=source_vision.patch_dim,
            vision_use_head=False,
        )

        output = HFGemma3Config(text_config=output_text, vision_config=output_vision)
        return output


@staticmethod
def _vision_projector_permute(ctx: io.TransformCTX, x):
    return torch.permute(x, (1, 0))


@io.state_transform(
    source_key=(
        "language_model.layers.*.self_attn.q_proj.weight",
        "language_model.layers.*.self_attn.k_proj.weight",
        "language_model.layers.*.self_attn.v_proj.weight",
    ),
    target_key="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
)
def _import_language_qkv(ctx: io.TransformCTX, q, k, v):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.language_transformer_config
    return import_qkv(
        q,
        k,
        v,
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=megatron_config.hidden_size,
        head_size=megatron_config.kv_channels,
    )


@io.state_transform(
    source_key="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
    target_key=(
        "language_model.layers.*.self_attn.q_proj.weight",
        "language_model.layers.*.self_attn.k_proj.weight",
        "language_model.layers.*.self_attn.v_proj.weight",
    ),
)
def _export_language_qkv(ctx: io.TransformCTX, qkv):
    # pylint: disable=C0115,C0116
    hf_config = ctx.target.config.text_config
    return export_qkv(
        qkv,
        head_num=hf_config.num_attention_heads,
        num_query_groups=hf_config.num_key_value_heads,
        heads_per_group=hf_config.num_attention_heads // hf_config.num_key_value_heads,
        hidden_size=hf_config.hidden_size,
        head_size=hf_config.head_dim,
    )


@io.state_transform(
    source_key=(
        "vision_tower.vision_model.encoder.layers.*.self_attn.q_proj.weight",
        "vision_tower.vision_model.encoder.layers.*.self_attn.k_proj.weight",
        "vision_tower.vision_model.encoder.layers.*.self_attn.v_proj.weight",
    ),
    target_key="vision_model.decoder.layers.*.self_attention.linear_qkv.weight",
)
def _import_vision_qkv(ctx: io.TransformCTX, q, k, v):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.vision_transformer_config
    return import_qkv(
        q,
        k,
        v,
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=megatron_config.hidden_size,
        head_size=megatron_config.kv_channels,
    )


@io.state_transform(
    source_key="vision_model.decoder.layers.*.self_attention.linear_qkv.weight",
    target_key=(
        "vision_tower.vision_model.encoder.layers.*.self_attn.q_proj.weight",
        "vision_tower.vision_model.encoder.layers.*.self_attn.k_proj.weight",
        "vision_tower.vision_model.encoder.layers.*.self_attn.v_proj.weight",
    ),
)
def _export_vision_qkv(ctx: io.TransformCTX, qkv):
    # pylint: disable=C0115,C0116
    hf_config = ctx.target.config.vision_config
    return export_qkv(
        qkv,
        head_num=hf_config.num_attention_heads,
        num_query_groups=hf_config.num_attention_heads,
        heads_per_group=1,
        hidden_size=hf_config.hidden_size,
        head_size=hf_config.hidden_size // hf_config.num_attention_heads,
    )


@io.state_transform(
    source_key=(
        "vision_tower.vision_model.encoder.layers.*.self_attn.q_proj.bias",
        "vision_tower.vision_model.encoder.layers.*.self_attn.k_proj.bias",
        "vision_tower.vision_model.encoder.layers.*.self_attn.v_proj.bias",
    ),
    target_key="vision_model.decoder.layers.*.self_attention.linear_qkv.bias",
)
def _import_vision_qkv_bias(ctx: io.TransformCTX, q_bias, k_bias, v_bias):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.vision_transformer_config
    return import_qkv(
        q_bias.unsqueeze(-1),
        k_bias.unsqueeze(-1),
        v_bias.unsqueeze(-1),
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=1,
        head_size=megatron_config.kv_channels,
    ).squeeze(-1)


@io.state_transform(
    source_key="vision_model.decoder.layers.*.self_attention.linear_qkv.bias",
    target_key=(
        "vision_tower.vision_model.encoder.layers.*.self_attn.q_proj.bias",
        "vision_tower.vision_model.encoder.layers.*.self_attn.k_proj.bias",
        "vision_tower.vision_model.encoder.layers.*.self_attn.v_proj.bias",
    ),
)
def _export_vision_qkv_bias(ctx: io.TransformCTX, qkv_bias):
    # pylint: disable=C0115,C0116
    hf_config = ctx.target.config.vision_config
    return export_qkv_bias(
        qkv_bias,
        head_num=hf_config.num_attention_heads,
        num_query_groups=hf_config.num_attention_heads,
        heads_per_group=1,
        head_size=hf_config.hidden_size // hf_config.num_attention_heads,
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from typing import List, Literal, Union

import pytest
import torch

from nemo.collections.asr.parts.utils.batched_beam_decoding_utils import (
    INIT_POINTER_VALUE,
    NON_EXISTENT_LABEL_VALUE,
    BatchedBeamHyps,
)
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis, NBestHypotheses


NestedFloatList = Union[float, List["NestedFloatList"]]  # recursive type alias


def assert_nested_lists_approx(
    actual: NestedFloatList, expected: NestedFloatList, rel_tol: float = 1e-4, abs_tol: float = 1e-4
) -> None:
    """
    Recursively asserts that two nested lists of floats are approximately equal
    within a given relative and absolute tolerance.
    """
    if isinstance(actual, list) and isinstance(expected, list):
        assert len(actual) == len(expected), f"Length mismatch: {len(actual)} != {len(expected)}"
        for act, exp in zip(actual, expected):
            assert_nested_lists_approx(act, exp, rel_tol, abs_tol)
    else:
        assert actual == pytest.approx(
            expected, rel=rel_tol, abs=abs_tol
        ), f"Values differ: actual={actual}, expected={expected}, rel_tol={rel_tol}, abs_tol={abs_tol}"


def assert_hyps_sequence_equal(
    actual: Union[List[int], torch.Tensor], expected: list[int], rel_tol: float = 1e-4, abs_tol: float = 1e-4
):
    """
    Asserts that two sequences of hypotheses are approximately equal.
    """
    if isinstance(actual, torch.Tensor):
        actual = actual.cpu().tolist()
    assert_nested_lists_approx(actual, expected, rel_tol, abs_tol)


def assert_hyps_timestamps_equal(
    actual: Union[List[int], torch.Tensor], expected: list[int], rel_tol: float = 1e-4, abs_tol: float = 1e-4
):
    """
    Asserts that two sequences of timestamp values are approximately equal.
    """
    if isinstance(actual, torch.Tensor):
        actual = actual.cpu().tolist()
    assert_nested_lists_approx(actual, expected, rel_tol, abs_tol)


DEVICES: List[torch.device] = [torch.device("cpu")]

if torch.cuda.is_available():
    DEVICES.append(torch.device("cuda"))

if hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
    DEVICES.append(torch.device("mps"))


class TestBatchedBeamHyps:
    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_instantiate(self, device: torch.device):
        _ = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=4, device=device, blank_index=1024)

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_size", [-1, 0])
    def test_rnnt_instantiate_incorrect_batch_size(self, batch_size: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=batch_size, beam_size=4, init_length=3, blank_index=1024)

    @pytest.mark.unit
    @pytest.mark.parametrize("beam_size", [-1, 0])
    def test_rnnt_instantiate_incorrect_beam_size(self, beam_size: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=2, beam_size=beam_size, init_length=3, blank_index=1024)

    @pytest.mark.unit
    @pytest.mark.parametrize("init_length", [-1, 0])
    def test_rnnt_instantiate_incorrect_init_length(self, init_length: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=1, beam_size=4, init_length=init_length, blank_index=1024)

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_add_results(self, device: torch.device):
        # batch of size 2, add label for first utterance
        hyps = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024)
        assert hyps._max_length == 1
        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )
        assert hyps._max_length == 2
        assert hyps.current_lengths_nb.tolist() == [[1, 0, 1], [1, 0, 0]]
        assert hyps.current_lengths_wb.tolist() == [[1, 1, 1], [1, 1, 1]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]])
        assert hyps.transcript_wb.tolist() == [
            [[0, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE], [1, NON_EXISTENT_LABEL_VALUE]],
            [[2, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE]],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, INIT_POINTER_VALUE], [1, INIT_POINTER_VALUE], [2, INIT_POINTER_VALUE]],
            [[0, INIT_POINTER_VALUE], [1, INIT_POINTER_VALUE], [2, INIT_POINTER_VALUE]],
        ]
        assert hyps.timestamps.tolist() == [
            [[0, 0], [1, 0], [0, 0]],
            [[0, 0], [1, 0], [1, 0]],
        ]
        assert hyps.next_timestamp.tolist() == [
            [0, 1, 0],
            [0, 1, 1],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_add_multiple_results(self, device: torch.device):
        hyps = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024)
        assert hyps._max_length == 1

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        assert hyps._max_length == 4
        assert hyps.current_lengths_nb.tolist() == [[2, 1, 0], [1, 0, 2]]
        assert hyps.current_lengths_wb.tolist() == [[2, 2, 2], [2, 2, 2]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1, 1024, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [2, 5, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 6, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [
                [0, 0, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [1, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [2, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
            ],
            [
                [0, 2, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [1, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [2, 0, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
            ],
        ]
        assert hyps.timestamps.tolist() == [
            [
                [0, 0, 0, 0],
                [1, 1, 0, 0],
                [0, 2, 0, 0],
            ],
            [
                [0, 1, 0, 0],
                [1, 2, 0, 0],
                [1, 0, 0, 0],
            ],
        ]
        assert hyps.next_timestamp.tolist() == [
            [0, 1, 2],
            [1, 2, 0],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_add_with_invalid_results(self, device: torch.device):
        hyps = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024)
        assert hyps._max_length == 1

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )

        assert hyps._max_length == 4
        assert hyps.current_lengths_nb.tolist() == [[1, 3, 1], [3, 1, 1]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, 7, NON_EXISTENT_LABEL_VALUE],
                [1, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [2, 5, 10, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 6, 9, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 1, INIT_POINTER_VALUE], [1, 1, 0, INIT_POINTER_VALUE], [2, 1, 2, INIT_POINTER_VALUE]],
            [[0, 2, 2, INIT_POINTER_VALUE], [1, 1, 0, INIT_POINTER_VALUE], [2, 0, 1, INIT_POINTER_VALUE]],
        ]
        assert hyps.timestamps.tolist() == [
            [
                [0, 0, 1, 0],
                [1, 1, 0, 0],
                [0, 2, 2, 0],
            ],
            [
                [0, 1, 0, 0],
                [1, 2, 1, 0],
                [1, 0, 2, 0],
            ],
        ]
        assert hyps.next_timestamp.tolist() == [
            [1, 0, 2],
            [0, 1, 2],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_instantiate(self, device: torch.device):
        _ = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=4, device=device, blank_index=1024, model_type='tdt'
        )

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_size", [-1, 0])
    def test_tdt_instantiate_incorrect_batch_size(self, batch_size: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=batch_size, beam_size=4, init_length=3, blank_index=1024, model_type='tdt')

    @pytest.mark.unit
    @pytest.mark.parametrize("beam_size", [-1, 0])
    def test_tdt_instantiate_incorrect_beam_size(self, beam_size: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=2, beam_size=beam_size, init_length=3, blank_index=1024, model_type='tdt')

    @pytest.mark.unit
    @pytest.mark.parametrize("init_length", [-1, 0])
    def test_tdt_instantiate_incorrect_init_length(self, init_length: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=1, beam_size=4, init_length=init_length, blank_index=1024, model_type='tdt')

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_add_results(self, device: torch.device):
        # batch of size 2, add label for first utterance
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='tdt'
        )
        assert hyps._max_length == 1
        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
            next_label_durations=torch.tensor([[0, 3, 1], [2, 3, 4]], device=device),
        )
        assert hyps._max_length == 2
        assert hyps.current_lengths_nb.tolist() == [[1, 0, 1], [1, 0, 0]]
        assert hyps.current_lengths_wb.tolist() == [[1, 1, 1], [1, 1, 1]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]])
        assert hyps.transcript_wb.tolist() == [
            [[0, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE], [1, NON_EXISTENT_LABEL_VALUE]],
            [[2, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE]],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, INIT_POINTER_VALUE], [1, INIT_POINTER_VALUE], [2, INIT_POINTER_VALUE]],
            [[0, INIT_POINTER_VALUE], [1, INIT_POINTER_VALUE], [2, INIT_POINTER_VALUE]],
        ]

        assert hyps.timestamps.tolist() == [[[0, 0], [3, 0], [1, 0]], [[2, 0], [3, 0], [4, 0]]]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_add_multiple_results(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='tdt'
        )
        assert hyps._max_length == 1

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
            next_label_durations=torch.tensor([[0, 3, 1], [2, 3, 4]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 4, 1], [0, 1, 1]], device=device),
        )

        assert hyps._max_length == 4
        assert hyps.current_lengths_nb.tolist() == [[2, 1, 0], [1, 0, 2]]
        assert hyps.current_lengths_wb.tolist() == [[2, 2, 2], [2, 2, 2]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1, 1024, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [2, 5, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 6, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [
                [0, 0, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [1, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [2, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
            ],
            [
                [0, 2, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [1, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [2, 0, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
            ],
        ]

        assert hyps.timestamps.tolist() == [
            [[0, 2, 0, 0], [3, 7, 0, 0], [1, 4, 0, 0]],
            [[2, 4, 0, 0], [3, 4, 0, 0], [4, 3, 0, 0]],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_add_with_invalid_results(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='tdt'
        )
        assert hyps._max_length == 1

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
            next_label_durations=torch.tensor([[0, 3, 1], [2, 3, 4]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 4, 1], [0, 1, 1]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 1, 3], [2, 1, 2]], device=device),
        )

        assert hyps._max_length == 4
        assert hyps.current_lengths_nb.tolist() == [[1, 3, 1], [3, 1, 1]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, 7, NON_EXISTENT_LABEL_VALUE],
                [1, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [2, 5, 10, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 6, 9, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 1, INIT_POINTER_VALUE], [1, 1, 0, INIT_POINTER_VALUE], [2, 1, 2, INIT_POINTER_VALUE]],
            [[0, 2, 2, INIT_POINTER_VALUE], [1, 1, 0, INIT_POINTER_VALUE], [2, 0, 1, INIT_POINTER_VALUE]],
        ]

        assert hyps.timestamps.tolist() == [
            [[0, 2, 7, 0], [3, 7, 3, 0], [1, 4, 7, 0]],
            [[2, 4, 5, 0], [3, 4, 4, 0], [4, 3, 6, 0]],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_ctc_instantiate(self, device: torch.device):
        _ = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=4, device=device, blank_index=1024, model_type='ctc'
        )

    @pytest.mark.unit
    @pytest.mark.parametrize("batch_size", [-1, 0])
    def test_ctc_instantiate_incorrect_batch_size(self, batch_size: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=batch_size, beam_size=4, init_length=3, blank_index=1024, model_type='ctc')

    @pytest.mark.unit
    @pytest.mark.parametrize("beam_size", [-1, 0])
    def test_ctc_instantiate_incorrect_beam_size(self, beam_size: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=2, beam_size=beam_size, init_length=3, blank_index=1024, model_type='ctc')

    @pytest.mark.unit
    @pytest.mark.parametrize("init_length", [-1, 0])
    def test_ctc_instantiate_incorrect_init_length(self, init_length: Literal[-1] | Literal[0]):
        with pytest.raises(ValueError):
            _ = BatchedBeamHyps(batch_size=1, beam_size=4, init_length=init_length, blank_index=1024)

    @pytest.mark.unit
    @pytest.mark.parametrize("y", [torch.tensor([1, 1024, 1024, 2, 2, 1024, 2, 3, 3, 1024, 3, 2, 2, 2])])
    def test_ctc_create_fold_consecutive_mask(self, y: torch.Tensor):
        batched_hyps = BatchedBeamHyps(batch_size=1, beam_size=4, init_length=30, blank_index=1024, model_type='ctc')
        mask = batched_hyps._create_fold_consecutive_mask(transcript=y)

        assert y[mask].tolist() == [1, 2, 2, 3, 3, 2]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_ctc_add_results(self, device: torch.device):
        # batch of size 2, add label for first utterance
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='ctc'
        )
        assert hyps._max_length == 1
        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )
        assert hyps._max_length == 2
        assert hyps.current_lengths_nb.tolist() == [[1, 0, 1], [1, 0, 0]]
        assert hyps.current_lengths_wb.tolist() == [[1, 1, 1], [1, 1, 1]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]])
        assert hyps.transcript_wb.tolist() == [
            [[0, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE], [1, NON_EXISTENT_LABEL_VALUE]],
            [[2, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE], [1024, NON_EXISTENT_LABEL_VALUE]],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, INIT_POINTER_VALUE], [1, INIT_POINTER_VALUE], [2, INIT_POINTER_VALUE]],
            [[0, INIT_POINTER_VALUE], [1, INIT_POINTER_VALUE], [2, INIT_POINTER_VALUE]],
        ]
        assert hyps.timestamps.tolist() == [
            [[0, 1], [0, 1], [0, 1]],
            [[0, 1], [0, 1], [0, 1]],
        ]
        assert hyps.last_label.tolist() == [
            [0, 1024, 1],
            [2, 1024, 1024],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_add_multiple_results(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='ctc'
        )
        assert hyps._max_length == 1

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        assert hyps._max_length == 4
        assert hyps.current_lengths_nb.tolist() == [[2, 1, 0], [1, 0, 2]]
        assert hyps.current_lengths_wb.tolist() == [[2, 2, 2], [2, 2, 2]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1, 1024, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [2, 5, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
                [1024, 6, NON_EXISTENT_LABEL_VALUE, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [
                [0, 0, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [1, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [2, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
            ],
            [
                [0, 2, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [1, 1, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
                [2, 0, INIT_POINTER_VALUE, INIT_POINTER_VALUE],
            ],
        ]
        assert hyps.timestamps.tolist() == [
            [
                [0, 1, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 2, 3],
            ],
            [
                [0, 1, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 2, 3],
            ],
        ]
        assert hyps.last_label.tolist() == [[3, 4, 1024], [5, 1024, 6]]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_add_with_invalid_results(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='ctc'
        )
        assert hyps._max_length == 1

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )

        assert hyps._max_length == 4
        assert hyps.current_lengths_nb.tolist() == [[1, 3, 1], [3, 1, 1]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, 7, NON_EXISTENT_LABEL_VALUE],
                [1, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [2, 5, 10, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 6, 9, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 1, INIT_POINTER_VALUE], [1, 1, 0, INIT_POINTER_VALUE], [2, 1, 2, INIT_POINTER_VALUE]],
            [[0, 2, 2, INIT_POINTER_VALUE], [1, 1, 0, INIT_POINTER_VALUE], [2, 0, 1, INIT_POINTER_VALUE]],
        ]
        assert hyps.timestamps.tolist() == [
            [
                [0, 1, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 2, 3],
            ],
            [
                [0, 1, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 2, 3],
            ],
        ]
        assert hyps.last_label.tolist() == [
            [4, 7, 8],
            [10, 5, 9],
        ]


class TestConvertToHypotheses:
    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_flatten_sort(self, device: torch.device):
        hyps = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024)

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )
        hyps.flatten_sort_(score_norm=False)

        assert hyps.current_lengths_nb.tolist() == [[3, 1, 1], [1, 1, 3]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.4, 0.35, 0.1], [0.6, 0.55, 0.4]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, 7, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [1024, 1024, 9, NON_EXISTENT_LABEL_VALUE],
                [1024, 5, -1, NON_EXISTENT_LABEL_VALUE],
                [2, 6, 10, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
        ]
        assert hyps.timestamps.tolist() == [
            [
                [0, 0, 0, 0],
                [1, 1, 1, 0],
                [1, 2, 2, 0],
            ],
            [
                [1, 2, 2, 0],
                [1, 1, 1, 0],
                [0, 0, 0, 0],
            ],
        ]
        assert hyps.next_timestamp.tolist() == [
            [0, 1, 2],
            [2, 1, 0],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_flatten_sort_norm(self, device: torch.device):
        hyps = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024)

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )

        hyps.flatten_sort_(score_norm=True)

        assert hyps.current_lengths_nb.tolist() == [[1, 3, 1], [1, 1, 3]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.35, 0.4, 0.1], [0.6, 0.55, 0.4]])
        assert hyps.transcript_wb.tolist() == [
            [
                [1024, 4, -1, NON_EXISTENT_LABEL_VALUE],
                [0, 3, 7, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [1024, 1024, 9, NON_EXISTENT_LABEL_VALUE],
                [1024, 5, -1, NON_EXISTENT_LABEL_VALUE],
                [2, 6, 10, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
        ]
        assert hyps.timestamps.tolist() == [
            [
                [1, 1, 1, 0],
                [0, 0, 0, 0],
                [1, 2, 2, 0],
            ],
            [
                [1, 2, 2, 0],
                [1, 1, 1, 0],
                [0, 0, 0, 0],
            ],
        ]
        assert hyps.next_timestamp.tolist() == [
            [1, 0, 2],
            [2, 1, 0],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_to_hyps_list(self, device: torch.device):
        hyps = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024)

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.4, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hypotheses = hyps.to_hyps_list(score_norm=False)

        assert type(hypotheses) == list
        assert type(hypotheses[0]) == Hypothesis
        assert type(hypotheses[1]) == Hypothesis

        assert len(hypotheses) == 2

        assert_hyps_sequence_equal(hypotheses[0].y_sequence, [0, 3, 7])
        assert_hyps_sequence_equal(hypotheses[1].y_sequence, [9])

        assert_hyps_timestamps_equal(hypotheses[0].timestamp, [0, 0, 0])
        assert_hyps_timestamps_equal(hypotheses[1].timestamp, [2])

        assert hypotheses[0].score == pytest.approx(0.4)
        assert hypotheses[1].score == pytest.approx(0.6)

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_rnnt_to_nbest_hyps_list(self, device: torch.device):
        hyps = BatchedBeamHyps(batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024)

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )

        hypotheses = hyps.to_nbest_hyps_list(score_norm=False)

        assert type(hypotheses) == list
        assert type(hypotheses[0]) == NBestHypotheses
        assert type(hypotheses[1]) == NBestHypotheses

        assert len(hypotheses) == 2
        assert len(hypotheses[0].n_best_hypotheses) == 3
        assert len(hypotheses[1].n_best_hypotheses) == 3

        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[0].y_sequence, [0, 3, 7])
        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[1].y_sequence, [4])
        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[2].y_sequence, [8])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[0].y_sequence, [9])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[1].y_sequence, [5])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[2].y_sequence, [2, 6, 10])

        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[0].timestamp, [0, 0, 0])
        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[1].timestamp, [1])
        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[2].timestamp, [2])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[0].timestamp, [2])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[1].timestamp, [1])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[2].timestamp, [0, 0, 0])

        assert hypotheses[0].n_best_hypotheses[0].score == pytest.approx(0.4)
        assert hypotheses[0].n_best_hypotheses[1].score == pytest.approx(0.35)
        assert hypotheses[0].n_best_hypotheses[2].score == pytest.approx(0.1)
        assert hypotheses[1].n_best_hypotheses[0].score == pytest.approx(0.6)
        assert hypotheses[1].n_best_hypotheses[1].score == pytest.approx(0.55)
        assert hypotheses[1].n_best_hypotheses[2].score == pytest.approx(0.4)

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_flatten_sort(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='tdt'
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
            next_label_durations=torch.tensor([[0, 3, 1], [2, 3, 4]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 4, 1], [0, 1, 1]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 1, 3], [2, 1, 2]], device=device),
        )

        hyps.flatten_sort_(score_norm=False)

        assert hyps.current_lengths_nb.tolist() == [[3, 1, 1], [1, 1, 3]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.4, 0.35, 0.1], [0.6, 0.55, 0.4]])
        assert hyps.transcript_wb.tolist() == [
            [
                [0, 3, 7, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [1024, 1024, 9, NON_EXISTENT_LABEL_VALUE],
                [1024, 5, -1, NON_EXISTENT_LABEL_VALUE],
                [2, 6, 10, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
        ]

        assert hyps.timestamps.tolist() == [
            [[0, 2, 3, 0], [3, 7, 7, 0], [3, 4, 7, 0]],
            [[3, 4, 6, 0], [4, 4, 4, 0], [2, 3, 5, 0]],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_flatten_sort_norm(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='tdt'
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
            next_label_durations=torch.tensor([[0, 3, 1], [2, 3, 4]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 4, 1], [0, 0, 1]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.4, 0.1], [0.4, 0.5, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 1, 3], [2, 1, 2]], device=device),
        )

        hyps.flatten_sort_(score_norm=True)

        assert hyps.current_lengths_nb.tolist() == [[1, 3, 1], [1, 1, 3]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.3, 0.4, 0.1], [0.6, 0.5, 0.4]])
        assert hyps.transcript_wb.tolist() == [
            [
                [1024, 4, -1, NON_EXISTENT_LABEL_VALUE],
                [0, 3, 7, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [1024, 1024, 9, NON_EXISTENT_LABEL_VALUE],
                [1024, 5, -1, NON_EXISTENT_LABEL_VALUE],
                [2, 6, 10, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
        ]

        assert hyps.timestamps.tolist() == [
            [[3, 7, 7, 0], [0, 2, 3, 0], [3, 4, 7, 0]],
            [[3, 3, 5, 0], [4, 4, 4, 0], [2, 3, 5, 0]],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_to_hyps_list(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='tdt'
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
            next_label_durations=torch.tensor([[0, 3, 1], [2, 3, 4]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 4, 1], [0, 1, 1]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 1, 3], [2, 1, 2]], device=device),
        )

        hypotheses = hyps.to_hyps_list(score_norm=False)

        assert type(hypotheses) == list
        assert type(hypotheses[0]) == Hypothesis
        assert type(hypotheses[1]) == Hypothesis

        assert len(hypotheses) == 2

        assert_hyps_sequence_equal(hypotheses[0].y_sequence, [0, 3, 7])
        assert_hyps_sequence_equal(hypotheses[1].y_sequence, [9])

        assert_hyps_timestamps_equal(hypotheses[0].timestamp, [0, 2, 3])
        assert_hyps_timestamps_equal(hypotheses[1].timestamp, [6])

        assert hypotheses[0].score == pytest.approx(0.4)
        assert hypotheses[1].score == pytest.approx(0.6)

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_tdt_to_nbest_hyps_list(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='tdt'
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[0, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
            next_label_durations=torch.tensor([[0, 3, 1], [2, 3, 4]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 4, 1], [0, 1, 1]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [10, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
            next_label_durations=torch.tensor([[2, 1, 3], [2, 1, 2]], device=device),
        )

        hypotheses = hyps.to_nbest_hyps_list(score_norm=False)

        assert type(hypotheses) == list
        assert type(hypotheses[0]) == NBestHypotheses
        assert type(hypotheses[1]) == NBestHypotheses

        assert len(hypotheses) == 2
        assert len(hypotheses[0].n_best_hypotheses) == 3
        assert len(hypotheses[1].n_best_hypotheses) == 3

        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[0].y_sequence, [0, 3, 7])
        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[1].y_sequence, [4])
        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[2].y_sequence, [8])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[0].y_sequence, [9])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[1].y_sequence, [5])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[2].y_sequence, [2, 6, 10])

        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[0].timestamp, [0, 2, 3])
        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[1].timestamp, [7])
        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[2].timestamp, [7])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[0].timestamp, [6])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[1].timestamp, [4])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[2].timestamp, [2, 3, 5])

        assert hypotheses[0].n_best_hypotheses[0].score == pytest.approx(0.4)
        assert hypotheses[0].n_best_hypotheses[1].score == pytest.approx(0.35)
        assert hypotheses[0].n_best_hypotheses[2].score == pytest.approx(0.1)
        assert hypotheses[1].n_best_hypotheses[0].score == pytest.approx(0.6)
        assert hypotheses[1].n_best_hypotheses[1].score == pytest.approx(0.55)
        assert hypotheses[1].n_best_hypotheses[2].score == pytest.approx(0.4)

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_ctc_flatten_sort(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='ctc'
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[3, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [2, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )
        hyps.flatten_sort_(score_norm=False)

        assert hyps.current_lengths_nb.tolist() == [[2, 1, 1], [1, 1, 3]]
        assert hyps.current_lengths_wb.tolist() == [[3, 3, 3], [3, 3, 3]]
        assert_nested_lists_approx(actual=hyps.scores.tolist(), expected=[[0.4, 0.35, 0.1], [0.6, 0.55, 0.4]])
        assert hyps.transcript_wb.tolist() == [
            [
                [3, 3, 7, NON_EXISTENT_LABEL_VALUE],
                [1024, 4, -1, NON_EXISTENT_LABEL_VALUE],
                [1024, 1024, 8, NON_EXISTENT_LABEL_VALUE],
            ],
            [
                [1024, 1024, 9, NON_EXISTENT_LABEL_VALUE],
                [1024, 5, -1, NON_EXISTENT_LABEL_VALUE],
                [2, 6, 2, NON_EXISTENT_LABEL_VALUE],
            ],
        ]
        assert hyps.transcript_wb_prev_ptr.tolist() == [
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
            [[0, 0, 0, INIT_POINTER_VALUE], [1, 1, 1, INIT_POINTER_VALUE], [2, 2, 2, INIT_POINTER_VALUE]],
        ]
        assert hyps.timestamps.tolist() == [
            [
                [0, 1, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 2, 3],
            ],
            [
                [0, 1, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 2, 3],
            ],
        ]
        assert hyps.last_label.tolist() == [
            [7, 4, 8],
            [9, 5, 2],
        ]

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_ctc_to_hyps_list(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='ctc'
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[3, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [2, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )

        hypotheses = hyps.to_hyps_list(score_norm=False)

        assert type(hypotheses) == list
        assert type(hypotheses[0]) == Hypothesis
        assert type(hypotheses[1]) == Hypothesis

        assert len(hypotheses) == 2

        assert_hyps_sequence_equal(hypotheses[0].y_sequence, [3, 7])
        assert_hyps_sequence_equal(hypotheses[1].y_sequence, [9])

        assert_hyps_timestamps_equal(hypotheses[0].timestamp, [0, 2])
        assert_hyps_timestamps_equal(hypotheses[1].timestamp, [2])

        assert hypotheses[0].score == pytest.approx(0.4)
        assert hypotheses[1].score == pytest.approx(0.6)

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_ctc_to_nbest_hyps_list(self, device: torch.device):
        hyps = BatchedBeamHyps(
            batch_size=2, beam_size=3, init_length=1, device=device, blank_index=1024, model_type='ctc'
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 2], [0, 1, 2]], device=device),
            next_labels=torch.tensor([[3, 1024, 1], [2, 1024, 1024]], device=device),
            next_hyps_prob=torch.tensor([[0.5, 0.6, 0.8], [0.1, 0.2, 0.3]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[0, 1, 1], [2, 1, 0]], device=device),
            next_labels=torch.tensor([[3, 4, 1024], [5, 1024, 6]], device=device),
            next_hyps_prob=torch.tensor([[0.3, 0.2, 0.1], [0.4, 0.5, 0.6]], device=device),
        )

        hyps.add_results_(
            next_indices=torch.tensor([[1, 0, 2], [2, 0, 1]], device=device),
            next_labels=torch.tensor([[-1, 7, 8], [2, -1, 9]], device=device),
            next_hyps_prob=torch.tensor([[0.35, 0.4, 0.1], [0.4, 0.55, 0.6]], device=device),
        )

        hypotheses = hyps.to_nbest_hyps_list(score_norm=False)

        assert type(hypotheses) == list
        assert type(hypotheses[0]) == NBestHypotheses
        assert type(hypotheses[1]) == NBestHypotheses

        assert len(hypotheses) == 2
        assert len(hypotheses[0].n_best_hypotheses) == 3
        assert len(hypotheses[1].n_best_hypotheses) == 3

        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[0].y_sequence, [3, 7])
        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[1].y_sequence, [4])
        assert_hyps_sequence_equal(hypotheses[0].n_best_hypotheses[2].y_sequence, [8])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[0].y_sequence, [9])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[1].y_sequence, [5])
        assert_hyps_sequence_equal(hypotheses[1].n_best_hypotheses[2].y_sequence, [2, 6, 2])

        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[0].timestamp, [0, 2])
        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[1].timestamp, [1])
        assert_hyps_timestamps_equal(hypotheses[0].n_best_hypotheses[2].timestamp, [2])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[0].timestamp, [2])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[1].timestamp, [1])
        assert_hyps_timestamps_equal(hypotheses[1].n_best_hypotheses[2].timestamp, [0, 1, 2])

        assert hypotheses[0].n_best_hypotheses[0].score == pytest.approx(0.4)
        assert hypotheses[0].n_best_hypotheses[1].score == pytest.approx(0.35)
        assert hypotheses[0].n_best_hypotheses[2].score == pytest.approx(0.1)
        assert hypotheses[1].n_best_hypotheses[0].score == pytest.approx(0.6)
        assert hypotheses[1].n_best_hypotheses[1].score == pytest.approx(0.55)
        assert hypotheses[1].n_best_hypotheses[2].score == pytest.approx(0.4)



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os
import tempfile
from typing import List, Optional, Union

import torch
from lightning.pytorch import Trainer
from omegaconf import DictConfig, OmegaConf, open_dict
from tqdm.auto import tqdm

from nemo.collections.asr.losses.ctc import CTCLoss
from nemo.collections.asr.metrics.wer import WER
from nemo.collections.asr.parts.mixins import ASRBPEMixin, InterCTCMixin
from nemo.collections.asr.parts.preprocessing.segment import ChannelSelectorType
from nemo.collections.asr.parts.submodules.ctc_decoding import CTCDecoding, CTCDecodingConfig
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis
from nemo.collections.multimodal.speech_cv.models.visual_rnnt_models import VisualEncDecRNNTModel
from nemo.core.classes.common import PretrainedModelInfo
from nemo.core.classes.mixins import AccessMixin
from nemo.utils import logging, model_utils


class VisualEncDecHybridRNNTCTCModel(VisualEncDecRNNTModel, ASRBPEMixin, InterCTCMixin):
    """Base class for hybrid RNNT/CTC models."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        cfg = model_utils.convert_model_config_to_dict_config(cfg)
        cfg = model_utils.maybe_update_config_version(cfg)
        super().__init__(cfg=cfg, trainer=trainer)

        if 'aux_ctc' not in self.cfg:
            raise ValueError(
                "The config need to have a section for the CTC decoder named as aux_ctc for Hybrid models."
            )
        with open_dict(self.cfg.aux_ctc):
            if "feat_in" not in self.cfg.aux_ctc.decoder or (
                not self.cfg.aux_ctc.decoder.feat_in and hasattr(self.encoder, '_feat_out')
            ):
                self.cfg.aux_ctc.decoder.feat_in = self.encoder._feat_out
            if "feat_in" not in self.cfg.aux_ctc.decoder or not self.cfg.aux_ctc.decoder.feat_in:
                raise ValueError("param feat_in of the decoder's config is not set!")

            if self.cfg.aux_ctc.decoder.num_classes < 1 and self.cfg.aux_ctc.decoder.vocabulary is not None:
                logging.info(
                    "\nReplacing placeholder number of classes ({}) with actual number of classes - {}".format(
                        self.cfg.aux_ctc.decoder.num_classes, len(self.cfg.aux_ctc.decoder.vocabulary)
                    )
                )
                self.cfg.aux_ctc.decoder["num_classes"] = len(self.cfg.aux_ctc.decoder.vocabulary)

        self.ctc_decoder = VisualEncDecHybridRNNTCTCModel.from_config_dict(self.cfg.aux_ctc.decoder)
        self.ctc_loss_weight = self.cfg.aux_ctc.get("ctc_loss_weight", 0.5)

        self.ctc_loss = CTCLoss(
            num_classes=self.ctc_decoder.num_classes_with_blank - 1,
            zero_infinity=True,
            reduction=self.cfg.aux_ctc.get("ctc_reduction", "mean_batch"),
        )

        ctc_decoding_cfg = self.cfg.aux_ctc.get('decoding', None)
        if ctc_decoding_cfg is None:
            ctc_decoding_cfg = OmegaConf.structured(CTCDecodingConfig)
            with open_dict(self.cfg.aux_ctc):
                self.cfg.aux_ctc.decoding = ctc_decoding_cfg

        self.ctc_decoding = CTCDecoding(self.cfg.aux_ctc.decoding, vocabulary=self.ctc_decoder.vocabulary)
        self.ctc_wer = WER(
            decoding=self.ctc_decoding,
            use_cer=self.cfg.aux_ctc.get('use_cer', False),
            dist_sync_on_step=True,
            log_prediction=self.cfg.get("log_prediction", False),
        )

        # setting the RNNT decoder as the default one
        self.use_rnnt_decoder = True

        # setting up interCTC loss (from InterCTCMixin)
        self.setup_interctc(decoder_name='decoder', loss_name='loss', wer_name='_wer')

    @torch.no_grad()
    def transcribe(
        self,
        paths2video_files: List[str],
        batch_size: int = 4,
        return_hypotheses: bool = False,
        partial_hypothesis: Optional[List['Hypothesis']] = None,
        num_workers: int = 0,
        channel_selector: Optional[ChannelSelectorType] = None,
    ) -> Union[List['Hypothesis'], Optional[List['Hypothesis']]]:
        """
        Uses greedy decoding to transcribe video files. Use this method for debugging and prototyping.

        Args:

        paths2video_files: (a list) of paths to video files.
        batch_size: (int) batch size to use during inference. \
        Bigger will result in better throughput performance but would use more memory.
            return_hypotheses: (bool) Either return hypotheses or text
        With hypotheses can do some postprocessing like getting timestamp or rescoring
            num_workers: (int) number of workers for DataLoader
            channel_selector (int | Iterable[int] | str): select a single channel or a subset of channels 
            from multi-channel audio. If set to `'average'`, it performs averaging across channels.
            Disabled if set to `None`. Defaults to `None`. Uses zero-based indexing.

        Returns:
            Returns a list of greedy transcript Hypothesis or list of all Hypothesis
        """
        if self.use_rnnt_decoder:
            return super().transcribe(
                paths2video_files=paths2video_files,
                batch_size=batch_size,
                return_hypotheses=return_hypotheses,
                partial_hypothesis=partial_hypothesis,
                num_workers=num_workers,
                channel_selector=channel_selector,
            )

        if paths2video_files is None or len(paths2video_files) == 0:
            return {}
        # We will store transcriptions here
        hypotheses = []
        # Model's mode and device
        mode = self.training
        device = next(self.parameters()).device

        if num_workers is None:
            num_workers = min(batch_size, os.cpu_count() - 1)

        try:

            # Switch model to evaluation mode
            self.eval()
            # Freeze the visual front-end, encoder and decoder modules
            self.video_front_end.freeze()
            self.encoder.freeze()
            self.decoder.freeze()
            self.joint.freeze()
            if hasattr(self, 'ctc_decoder'):
                self.ctc_decoder.freeze()

            logging_level = logging.get_verbosity()
            logging.set_verbosity(logging.WARNING)
            # Work in tmp directory - will store manifest file there
            with tempfile.TemporaryDirectory() as tmpdir:
                with open(os.path.join(tmpdir, 'manifest.json'), 'w', encoding='utf-8') as fp:
                    for video_file in paths2video_files:
                        entry = {'video_filepath': video_file, 'duration': 100000, 'text': ''}
                        fp.write(json.dumps(entry) + '\n')

                config = {
                    'paths2video_files': paths2video_files,
                    'batch_size': batch_size,
                    'temp_dir': tmpdir,
                    'num_workers': num_workers,
                    'channel_selector': channel_selector,
                }

                temporary_datalayer = self._setup_transcribe_dataloader(config)
                for test_batch in tqdm(temporary_datalayer, desc="Transcribing"):
                    encoded, encoded_len = self.forward(
                        input_signal=test_batch[0].to(device), input_signal_length=test_batch[1].to(device)
                    )

                    logits = self.ctc_decoder(encoder_output=encoded)
                    best_hyp = self.ctc_decoding.ctc_decoder_predictions_tensor(
                        logits,
                        encoded_len,
                        return_hypotheses=return_hypotheses,
                    )
                    if return_hypotheses:
                        # dump log probs per file
                        for idx in range(logits.shape[0]):
                            best_hyp[idx].y_sequence = logits[idx][: encoded_len[idx]]
                            if best_hyp[idx].alignments is None:
                                best_hyp[idx].alignments = best_hyp[idx].y_sequence
                    del logits

                    hypotheses += best_hyp

                    del encoded
                    del test_batch
        finally:
            # set mode back to its original value
            self.train(mode=mode)

            logging.set_verbosity(logging_level)
            if mode is True:
                self.video_front_end.unfreeze()
                self.encoder.unfreeze()
                self.decoder.unfreeze()
                self.joint.unfreeze()
                if hasattr(self, 'ctc_decoder'):
                    self.ctc_decoder.unfreeze()
        return hypotheses

    def change_vocabulary(
        self,
        new_vocabulary: List[str],
        decoding_cfg: Optional[DictConfig] = None,
        ctc_decoding_cfg: Optional[DictConfig] = None,
    ):
        """
        Changes vocabulary used during RNNT decoding process. Use this method when fine-tuning a pre-trained model.
        This method changes only decoder and leaves encoder and pre-processing modules unchanged. For example, you
        would use it if you want to use pretrained encoder when fine-tuning on data in another language, or when
        you'd need model to learn capitalization, punctuation and/or special characters.

        Args:
            new_vocabulary: list with new vocabulary. Must contain at least 2 elements. Typically, \
                this is target alphabet.
            decoding_cfg: A config for the decoder, which is optional. If the decoding type
                needs to be changed (from say Greedy to Beam decoding etc), the config can be passed here.
            ctc_decoding_cfg: A config for CTC decoding, which is optional and can be used to change decoding type.

        Returns: None

        """
        super().change_vocabulary(new_vocabulary=new_vocabulary, decoding_cfg=decoding_cfg)

        # set up the new tokenizer for the CTC decoder
        if hasattr(self, 'ctc_decoder'):
            if self.ctc_decoder.vocabulary == new_vocabulary:
                logging.warning(
                    f"Old {self.ctc_decoder.vocabulary} and new {new_vocabulary} match. Not changing anything."
                )
            else:
                if new_vocabulary is None or len(new_vocabulary) == 0:
                    raise ValueError(f'New vocabulary must be non-empty list of chars. But I got: {new_vocabulary}')
                decoder_config = self.ctc_decoder.to_config_dict()
                new_decoder_config = copy.deepcopy(decoder_config)
                new_decoder_config['vocabulary'] = new_vocabulary
                new_decoder_config['num_classes'] = len(new_vocabulary)

                del self.ctc_decoder
                self.ctc_decoder = VisualEncDecHybridRNNTCTCModel.from_config_dict(new_decoder_config)
                del self.ctc_loss
                self.ctc_loss = CTCLoss(
                    num_classes=self.ctc_decoder.num_classes_with_blank - 1,
                    zero_infinity=True,
                    reduction=self.cfg.aux_ctc.get("ctc_reduction", "mean_batch"),
                )

                if ctc_decoding_cfg is None:
                    # Assume same decoding config as before
                    logging.info("No `ctc_decoding_cfg` passed when changing decoding strategy, using internal config")
                    ctc_decoding_cfg = self.cfg.aux_ctc.decoding

                # Assert the decoding config with all hyper parameters
                ctc_decoding_cls = OmegaConf.structured(CTCDecodingConfig)
                ctc_decoding_cls = OmegaConf.create(OmegaConf.to_container(ctc_decoding_cls))
                ctc_decoding_cfg = OmegaConf.merge(ctc_decoding_cls, ctc_decoding_cfg)

                self.ctc_decoding = CTCDecoding(decoding_cfg=ctc_decoding_cfg, vocabulary=self.ctc_decoder.vocabulary)

                self.ctc_wer = WER(
                    decoding=self.ctc_decoding,
                    use_cer=self.ctc_wer.use_cer,
                    log_prediction=self.ctc_wer.log_prediction,
                    dist_sync_on_step=True,
                )

                # Update config
                with open_dict(self.cfg.aux_ctc):
                    self.cfg.aux_ctc.decoding = ctc_decoding_cfg

                with open_dict(self.cfg.aux_ctc):
                    self.cfg.aux_ctc.decoder = new_decoder_config

                ds_keys = ['train_ds', 'validation_ds', 'test_ds']
                for key in ds_keys:
                    if key in self.cfg:
                        with open_dict(self.cfg[key]):
                            self.cfg[key]['labels'] = OmegaConf.create(new_vocabulary)

                logging.info(f"Changed the tokenizer of the CTC decoder to {self.ctc_decoder.vocabulary} vocabulary.")

    def change_decoding_strategy(self, decoding_cfg: DictConfig, decoder_type: str = None):
        """
        Changes decoding strategy used during RNNT decoding process.

        Args:
            decoding_cfg: A config for the decoder, which is optional. If the decoding type
                needs to be changed (from say Greedy to Beam decoding etc), the config can be passed here.
            decoder_type: (str) Can be set to 'rnnt' or 'ctc' to switch between appropriate decoder in a
                model having RNN-T and CTC decoders. Defaults to None, in which case RNN-T decoder is
                used. If set to 'ctc', it raises error if 'ctc_decoder' is not an attribute of the model.
        """
        if decoder_type is None or decoder_type == 'rnnt':
            self.use_rnnt_decoder = True
            return super().change_decoding_strategy(decoding_cfg=decoding_cfg)

        assert decoder_type == 'ctc' and hasattr(self, 'ctc_decoder')
        if decoding_cfg is None:
            # Assume same decoding config as before
            logging.info("No `decoding_cfg` passed when changing decoding strategy, using internal config")
            decoding_cfg = self.cfg.aux_ctc.decoding

        # Assert the decoding config with all hyper parameters
        decoding_cls = OmegaConf.structured(CTCDecodingConfig)
        decoding_cls = OmegaConf.create(OmegaConf.to_container(decoding_cls))
        decoding_cfg = OmegaConf.merge(decoding_cls, decoding_cfg)

        self.ctc_decoding = CTCDecoding(decoding_cfg=decoding_cfg, vocabulary=self.ctc_decoder.vocabulary)

        self.ctc_wer = WER(
            decoding=self.ctc_decoding,
            use_cer=self.ctc_wer.use_cer,
            log_prediction=self.ctc_wer.log_prediction,
            dist_sync_on_step=True,
        )

        # Update config
        with open_dict(self.cfg.aux_ctc):
            self.cfg.aux_ctc.decoding = decoding_cfg

        self.use_rnnt_decoder = False
        logging.info(f"Changed decoding strategy to \n{OmegaConf.to_yaml(self.cfg.aux_ctc.decoding)}")

    # PTL-specific methods
    def training_step(self, batch, batch_nb):
        # Reset access registry
        if AccessMixin.is_access_enabled(getattr(self, "model_guid", None)):
            AccessMixin.reset_registry(self)

        if self.is_interctc_enabled():
            AccessMixin.set_access_enabled(access_enabled=True, guid=self.model_guid)

        signal, signal_len, transcript, transcript_len = batch

        # forward() only performs encoder forward
        encoded, encoded_len = self.forward(input_signal=signal, input_signal_length=signal_len)
        del signal

        # During training, loss must be computed, so decoder forward is necessary
        decoder, target_length, states = self.decoder(targets=transcript, target_length=transcript_len)

        if hasattr(self, '_trainer') and self._trainer is not None:
            log_every_n_steps = self._trainer.log_every_n_steps
            sample_id = self._trainer.global_step
        else:
            log_every_n_steps = 1
            sample_id = batch_nb

        # If fused Joint-Loss-WER is not used
        if not self.joint.fuse_loss_wer:
            # Compute full joint and loss
            joint = self.joint(encoder_outputs=encoded, decoder_outputs=decoder)
            loss_value = self.loss(
                log_probs=joint, targets=transcript, input_lengths=encoded_len, target_lengths=target_length
            )

            # Add auxiliary losses, if registered
            loss_value = self.add_auxiliary_losses(loss_value)

            # Reset access registry
            # if AccessMixin.is_access_enabled():
            #    AccessMixin.reset_registry(self)

            tensorboard_logs = {
                'train_loss': loss_value,
                'learning_rate': self._optimizer.param_groups[0]['lr'],
                'global_step': torch.tensor(self.trainer.global_step, dtype=torch.float32),
            }

            if (sample_id + 1) % log_every_n_steps == 0:
                self.wer.update(
                    predictions=encoded,
                    predictions_lengths=encoded_len,
                    targets=transcript,
                    targets_lengths=transcript_len,
                )
                _, scores, words = self.wer.compute()
                self.wer.reset()
                tensorboard_logs.update({'training_batch_wer': scores.float() / words})

        else:
            # If fused Joint-Loss-WER is used
            if (sample_id + 1) % log_every_n_steps == 0:
                compute_wer = True
            else:
                compute_wer = False

            # Fused joint step
            loss_value, wer, _, _ = self.joint(
                encoder_outputs=encoded,
                decoder_outputs=decoder,
                encoder_lengths=encoded_len,
                transcripts=transcript,
                transcript_lengths=transcript_len,
                compute_wer=compute_wer,
            )

            # Add auxiliary losses, if registered
            loss_value = self.add_auxiliary_losses(loss_value)

            # Reset access registry
            # if AccessMixin.is_access_enabled():
            #    AccessMixin.reset_registry(self)

            tensorboard_logs = {
                'train_loss': loss_value,
                'learning_rate': self._optimizer.param_groups[0]['lr'],
                'global_step': torch.tensor(self.trainer.global_step, dtype=torch.float32),
            }

            if compute_wer:
                tensorboard_logs.update({'training_batch_wer': wer})

        if self.ctc_loss_weight > 0:
            log_probs = self.ctc_decoder(encoder_output=encoded)
            ctc_loss = self.ctc_loss(
                log_probs=log_probs, targets=transcript, input_lengths=encoded_len, target_lengths=transcript_len
            )

            # Add Interctc Losses
            ctc_loss, interctc_tensorboard_logs = self.add_interctc_losses(
                ctc_loss, transcript, transcript_len, compute_wer=((batch_nb + 1) % log_every_n_steps == 0)
            )
            tensorboard_logs.update(interctc_tensorboard_logs)

            tensorboard_logs['train_rnnt_loss'] = loss_value
            tensorboard_logs['train_ctc_loss'] = ctc_loss
            loss_value = (1 - self.ctc_loss_weight) * loss_value + self.ctc_loss_weight * ctc_loss
            tensorboard_logs['train_loss'] = loss_value
            if (sample_id + 1) % log_every_n_steps == 0:
                self.ctc_wer.update(
                    predictions=log_probs,
                    targets=transcript,
                    target_lengths=transcript_len,
                    predictions_lengths=encoded_len,
                )
                ctc_wer, _, _ = self.ctc_wer.compute()
                self.ctc_wer.reset()
                tensorboard_logs.update({'training_batch_wer_ctc': ctc_wer})

        # Reset access registry
        if AccessMixin.is_access_enabled(getattr(self, "model_guid", None)):
            AccessMixin.reset_registry(self)

        # Log items
        self.log_dict(tensorboard_logs)

        # Preserve batch acoustic model T and language model U parameters if normalizing
        if self._optim_normalize_joint_txu:
            self._optim_normalize_txu = [encoded_len.max(), transcript_len.max()]

        return {'loss': loss_value}

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        # TODO: add support for CTC decoding
        signal, signal_len, transcript, transcript_len, sample_id = batch

        # forward() only performs encoder forward
        encoded, encoded_len = self.forward(input_signal=signal, input_signal_length=signal_len)
        del signal

        best_hyp_text = self.decoding.rnnt_decoder_predictions_tensor(
            encoder_output=encoded, encoded_lengths=encoded_len, return_hypotheses=False
        )

        sample_id = sample_id.cpu().detach().numpy()
        return list(zip(sample_id, best_hyp_text))

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        if self.is_interctc_enabled():
            AccessMixin.set_access_enabled(access_enabled=True, guid=self.model_guid)

        signal, signal_len, transcript, transcript_len = batch

        # forward() only performs encoder forward
        encoded, encoded_len = self.forward(input_signal=signal, input_signal_length=signal_len)
        del signal

        tensorboard_logs = {}

        # If experimental fused Joint-Loss-WER is not used
        if not self.joint.fuse_loss_wer:
            if self.compute_eval_loss:
                decoder, target_length, states = self.decoder(targets=transcript, target_length=transcript_len)
                joint = self.joint(encoder_outputs=encoded, decoder_outputs=decoder)

                loss_value = self.loss(
                    log_probs=joint, targets=transcript, input_lengths=encoded_len, target_lengths=target_length
                )

                tensorboard_logs['val_loss'] = loss_value

            self.wer.update(
                predictions=encoded,
                predictions_lengths=encoded_len,
                targets=transcript,
                targets_lengths=transcript_len,
            )
            wer, wer_num, wer_denom = self.wer.compute()
            self.wer.reset()

            tensorboard_logs['val_wer_num'] = wer_num
            tensorboard_logs['val_wer_denom'] = wer_denom
            tensorboard_logs['val_wer'] = wer

        else:
            # If experimental fused Joint-Loss-WER is used
            compute_wer = True

            if self.compute_eval_loss:
                decoded, target_len, states = self.decoder(targets=transcript, target_length=transcript_len)
            else:
                decoded = None
                target_len = transcript_len

            # Fused joint step
            loss_value, wer, wer_num, wer_denom = self.joint(
                encoder_outputs=encoded,
                decoder_outputs=decoded,
                encoder_lengths=encoded_len,
                transcripts=transcript,
                transcript_lengths=target_len,
                compute_wer=compute_wer,
            )

            if loss_value is not None:
                tensorboard_logs['val_loss'] = loss_value

            tensorboard_logs['val_wer_num'] = wer_num
            tensorboard_logs['val_wer_denom'] = wer_denom
            tensorboard_logs['val_wer'] = wer

        log_probs = self.ctc_decoder(encoder_output=encoded)
        if self.compute_eval_loss:
            ctc_loss = self.ctc_loss(
                log_probs=log_probs, targets=transcript, input_lengths=encoded_len, target_lengths=transcript_len
            )

            # Add interCTC losses
            ctc_loss, interctc_tensorboard_logs = self.add_interctc_losses(
                ctc_loss,
                transcript,
                transcript_len,
                compute_wer=True,
                log_wer_num_denom=True,
                log_prefix="val_",
            )
            tensorboard_logs.update(interctc_tensorboard_logs)

            tensorboard_logs['val_ctc_loss'] = ctc_loss
            tensorboard_logs['val_rnnt_loss'] = loss_value
            loss_value = (1 - self.ctc_loss_weight) * loss_value + self.ctc_loss_weight * ctc_loss
            tensorboard_logs['val_loss'] = loss_value
        self.ctc_wer.update(
            predictions=log_probs,
            targets=transcript,
            target_lengths=transcript_len,
            predictions_lengths=encoded_len,
        )
        ctc_wer, ctc_wer_num, ctc_wer_denom = self.ctc_wer.compute()
        self.ctc_wer.reset()
        tensorboard_logs['val_wer_num_ctc'] = ctc_wer_num
        tensorboard_logs['val_wer_denom_ctc'] = ctc_wer_denom
        tensorboard_logs['val_wer_ctc'] = ctc_wer

        self.log('global_step', torch.tensor(self.trainer.global_step, dtype=torch.float32))

        # Reset access registry
        if AccessMixin.is_access_enabled(getattr(self, "model_guid", None)):
            AccessMixin.reset_registry(self)

        return tensorboard_logs

    def test_step(self, batch, batch_idx, dataloader_idx=0):
        logs = self.validation_step(batch, batch_idx, dataloader_idx=dataloader_idx)
        test_logs = {name.replace("val_", "test_"): value for name, value in logs.items()}
        if type(self.trainer.test_dataloaders) == list and len(self.trainer.test_dataloaders) > 1:
            self.test_step_outputs[dataloader_idx].append(test_logs)
        else:
            self.test_step_outputs.append(test_logs)
        return test_logs

    """
    def test_step(self, batch, batch_idx, dataloader_idx=0):
        logs = self.validation_step(batch, batch_idx, dataloader_idx=dataloader_idx)
        test_logs = {
            'test_wer_num': logs['val_wer_num'],
            'test_wer_denom': logs['val_wer_denom'],
            # 'test_wer': logs['val_wer'],
        }
        if 'val_loss' in logs:
            test_logs['test_loss'] = logs['val_loss']

        if self.ctc_loss_weight > 0:
            test_logs['test_wer_num_ctc'] = logs['val_wer_num_ctc']
            test_logs['test_wer_denom_ctc'] = logs['val_wer_denom_ctc']
            if 'val_ctc_loss' in logs:
                test_logs['test_ctc_loss'] = logs['val_ctc_loss']
            if 'val_rnnt_loss' in logs:
                test_logs['test_rnnt_loss'] = logs['val_rnnt_loss']

        return test_logs
    """

    def multi_validation_epoch_end(self, outputs, dataloader_idx: int = 0):
        if self.compute_eval_loss:
            val_loss_mean = torch.stack([x['val_loss'] for x in outputs]).mean()
            val_loss_log = {'val_loss': val_loss_mean}
        else:
            val_loss_log = {}
        wer_num = torch.stack([x['val_wer_num'] for x in outputs]).sum()
        wer_denom = torch.stack([x['val_wer_denom'] for x in outputs]).sum()
        tensorboard_logs = {**val_loss_log, 'val_wer': wer_num.float() / wer_denom}
        if self.ctc_loss_weight > 0:
            ctc_wer_num = torch.stack([x['val_wer_num_ctc'] for x in outputs]).sum()
            ctc_wer_denom = torch.stack([x['val_wer_denom_ctc'] for x in outputs]).sum()
            tensorboard_logs['val_wer_ctc'] = ctc_wer_num.float() / ctc_wer_denom

        metrics = {**val_loss_log, 'log': tensorboard_logs}
        self.finalize_interctc_metrics(metrics, outputs, prefix="val_")
        return metrics

    def multi_test_epoch_end(self, outputs, dataloader_idx: int = 0):
        if self.compute_eval_loss:
            test_loss_mean = torch.stack([x['test_loss'] for x in outputs]).mean()
            test_loss_log = {'test_loss': test_loss_mean}
        else:
            test_loss_log = {}
        wer_num = torch.stack([x['test_wer_num'] for x in outputs]).sum()
        wer_denom = torch.stack([x['test_wer_denom'] for x in outputs]).sum()
        tensorboard_logs = {**test_loss_log, 'test_wer': wer_num.float() / wer_denom}

        if self.ctc_loss_weight > 0:
            ctc_wer_num = torch.stack([x['test_wer_num_ctc'] for x in outputs]).sum()
            ctc_wer_denom = torch.stack([x['test_wer_denom_ctc'] for x in outputs]).sum()
            tensorboard_logs['test_wer_ctc'] = ctc_wer_num.float() / ctc_wer_denom

        metrics = {**test_loss_log, 'log': tensorboard_logs}
        self.finalize_interctc_metrics(metrics, outputs, prefix="test_")
        return metrics

    @classmethod
    def list_available_models(cls) -> Optional[PretrainedModelInfo]:
        """
        This method returns a list of pre-trained model which can be instantiated directly from NVIDIA's NGC cloud.

        Returns:
            List of available pre-trained models.
        """
        results = []
        return results



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

from nemo.collections.asr.parts.submodules import adapters as adapter_modules
from nemo.core.classes.mixins import adapter_mixin_strategies
from nemo.utils import config_utils


def _create_masks(att_mask, max_audio_length, padding_length):
    # pad_mask is the masking to be used to ignore paddings
    pad_mask = torch.arange(0, max_audio_length).expand(padding_length.size(0), -1) < padding_length.unsqueeze(-1)

    # pad_mask_for_att_mask is the mask which helps to ignore paddings
    pad_mask_for_att_mask = pad_mask.unsqueeze(1).repeat([1, max_audio_length, 1])
    pad_mask_for_att_mask = torch.logical_and(pad_mask_for_att_mask, pad_mask_for_att_mask.transpose(1, 2))
    # att_mask is the masking to be used by the MHA layers to ignore the tokens not supposed to be visible
    att_mask = att_mask[:, :max_audio_length, :max_audio_length]
    # paddings should also get ignored, so pad_mask_for_att_mask is used to ignore their corresponding scores
    att_mask = torch.logical_and(pad_mask_for_att_mask, att_mask.to(pad_mask_for_att_mask.device))

    pad_mask = ~pad_mask
    att_mask = ~att_mask

    return pad_mask, att_mask


def get_mask(lengths: torch.Tensor):
    max_seq_len = lengths.max()
    att_mask = torch.ones(1, max_seq_len, max_seq_len, dtype=torch.bool)

    pad_mask, att_mask = _create_masks(att_mask, max_seq_len, lengths)
    return pad_mask, att_mask


class TestASRAdapterModules:
    @pytest.mark.unit
    def test_mha_adapter_config(self):
        IGNORED_ARGS = ['_target_']

        result = config_utils.assert_dataclass_signature_match(
            adapter_modules.MultiHeadAttentionAdapter,
            adapter_modules.MultiHeadAttentionAdapterConfig,
            ignore_args=IGNORED_ARGS,
        )

        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None

    @pytest.mark.unit
    def test_relpos_mha_adapter_config(self):
        IGNORED_ARGS = ['_target_']

        result = config_utils.assert_dataclass_signature_match(
            adapter_modules.RelPositionMultiHeadAttentionAdapter,
            adapter_modules.RelPositionMultiHeadAttentionAdapterConfig,
            ignore_args=IGNORED_ARGS,
        )

        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None

    @pytest.mark.unit
    def test_abs_pos_encoding_adapter_config(self):
        IGNORED_ARGS = ['_target_']

        result = config_utils.assert_dataclass_signature_match(
            adapter_modules.PositionalEncodingAdapter,
            adapter_modules.PositionalEncodingAdapterConfig,
            ignore_args=IGNORED_ARGS,
        )

        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None

    @pytest.mark.unit
    def test_rel_pos_encoding_adapter_config(self):
        IGNORED_ARGS = ['_target_']

        result = config_utils.assert_dataclass_signature_match(
            adapter_modules.RelPositionalEncodingAdapter,
            adapter_modules.RelPositionalEncodingAdapterConfig,
            ignore_args=IGNORED_ARGS,
        )

        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None

    @pytest.mark.unit
    def test_transformer_mha_adapter_config(self):
        IGNORED_ARGS = ['_target_']

        result = config_utils.assert_dataclass_signature_match(
            adapter_modules.TransformerMultiHeadAttentionAdapter,
            adapter_modules.TransformerMultiHeadAttentionAdapterConfig,
            ignore_args=IGNORED_ARGS,
        )

        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None

    @pytest.mark.unit
    @pytest.mark.parametrize('n_head', [1, 2, 10])
    @pytest.mark.parametrize('proj_dim', [None, -1])
    def test_mha_adapter_init(self, n_head, proj_dim):
        torch.random.manual_seed(0)
        x = torch.randn(2, 32, 50)
        lengths = torch.randint(1, x.size(1), size=(x.size(0),))
        lengths[torch.randint(0, x.size(0), size=(1,))[0]] = x.size(1)

        adapter = adapter_modules.MultiHeadAttentionAdapter(
            n_head=n_head, n_feat=50, dropout_rate=0.0, proj_dim=proj_dim
        )

        pad_mask, att_mask = get_mask(lengths)

        with torch.no_grad():
            assert adapter.linear_out.weight.sum() == 0
            if hasattr(adapter.linear_out, 'bias') and adapter.linear_out.bias is not None:
                assert adapter.linear_out.bias.sum() == 0

            out = adapter(x, x, x, att_mask)
            assert out.sum().abs() <= 1e-8
            assert out.shape == x.shape

    @pytest.mark.unit
    @pytest.mark.parametrize('n_head', [1, 2, 10])
    @pytest.mark.parametrize('proj_dim', [None, -1])
    def test_relmha_adapter_init(self, n_head, proj_dim):
        torch.random.manual_seed(0)
        x = torch.randn(2, 32, 50)
        lengths = torch.randint(1, x.size(1), size=(x.size(0),))
        lengths[torch.randint(0, x.size(0), size=(1,))[0]] = x.size(1)

        adapter = adapter_modules.RelPositionMultiHeadAttentionAdapter(
            n_head=n_head, n_feat=50, dropout_rate=0.0, proj_dim=proj_dim
        )
        relpos_enc = adapter_modules.RelPositionalEncodingAdapter(d_model=50)

        pad_mask, att_mask = get_mask(lengths)
        relpos_enc.extend_pe(lengths.max(), device='cpu', dtype=torch.float32)

        with torch.no_grad():
            assert adapter.linear_out.weight.sum() == 0
            if hasattr(adapter.linear_out, 'bias') and adapter.linear_out.bias is not None:
                assert adapter.linear_out.bias.sum() == 0

            _, pos_emb = relpos_enc(x)
            out = adapter(x, x, x, att_mask, pos_emb)
            assert out.sum().abs() <= 1e-8
            assert out.shape == x.shape

    @pytest.mark.unit
    def test_relmha_adapter_with_torch_sdpa(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 32, 50)
        lengths = torch.randint(1, x.size(1), size=(x.size(0),))
        lengths[torch.randint(0, x.size(0), size=(1,))[0]] = x.size(1)

        adapter_torch_sdpa = adapter_modules.RelPositionMultiHeadAttentionAdapter(
            n_head=2, n_feat=50, dropout_rate=0.0, proj_dim=-1, use_pytorch_sdpa=True
        )
        adapter = adapter_modules.RelPositionMultiHeadAttentionAdapter(
            n_head=2, n_feat=50, dropout_rate=0.0, proj_dim=-1, use_pytorch_sdpa=False
        )
        # to dont reset linear_out parameters to zero
        adapter.linear_out = torch.nn.Linear(adapter.linear_out.in_features, adapter.linear_out.out_features)
        for original_param, sdpa_param in zip(adapter.parameters(), adapter_torch_sdpa.parameters()):
            sdpa_param.data.copy_(original_param.data)
        relpos_enc = adapter_modules.RelPositionalEncodingAdapter(d_model=50)

        pad_mask, att_mask = get_mask(lengths)
        relpos_enc.extend_pe(lengths.max(), device='cpu', dtype=torch.float32)

        with torch.no_grad():
            _, pos_emb = relpos_enc(x)
            out = adapter(x, x, x, att_mask, pos_emb)
            out_sdpa = adapter_torch_sdpa(x, x, x, att_mask, pos_emb)
            assert torch.allclose(out_sdpa, out, atol=1e-5)

    @pytest.mark.unit
    def test_mha_adapter_with_torch_sdpa(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 32, 50)
        lengths = torch.randint(1, x.size(1), size=(x.size(0),))
        lengths[torch.randint(0, x.size(0), size=(1,))[0]] = x.size(1)

        adapter_torch_sdpa = adapter_modules.MultiHeadAttentionAdapter(
            n_head=2, n_feat=50, dropout_rate=0.0, proj_dim=-1, use_pytorch_sdpa=True
        )
        adapter = adapter_modules.MultiHeadAttentionAdapter(
            n_head=2, n_feat=50, dropout_rate=0.0, proj_dim=-1, use_pytorch_sdpa=False
        )
        # to dont reset linear_out parameters to zero
        adapter.linear_out = torch.nn.Linear(adapter.linear_out.in_features, adapter.linear_out.out_features)

        for original_param, sdpa_param in zip(adapter.parameters(), adapter_torch_sdpa.parameters()):
            sdpa_param.data.copy_(original_param.data)

        pad_mask, att_mask = get_mask(lengths)
        with torch.no_grad():
            out = adapter(x, x, x, att_mask)
            out_sdpa = adapter_torch_sdpa(x, x, x, att_mask)
            assert torch.allclose(out_sdpa, out, atol=1e-5)

    @pytest.mark.unit
    def test_abspos_encoding_init(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 32, 50)
        lengths = torch.randint(1, x.size(1), size=(x.size(0),))
        lengths[torch.randint(0, x.size(0), size=(1,))[0]] = x.size(1)

        relpos_enc = adapter_modules.PositionalEncodingAdapter(d_model=50)

        relpos_enc.extend_pe(lengths.max(), device='cpu', dtype=torch.float32)

        with torch.no_grad():
            out, pos_emb = relpos_enc(x)
            assert (out - pos_emb - x).sum().abs() <= 1e-5
            assert out.shape == x.shape

    @pytest.mark.unit
    def test_relpos_encoding_init(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 32, 50)
        lengths = torch.randint(1, x.size(1), size=(x.size(0),))
        lengths[torch.randint(0, x.size(0), size=(1,))[0]] = x.size(1)

        relpos_enc = adapter_modules.RelPositionalEncodingAdapter(d_model=50)

        relpos_enc.extend_pe(lengths.max(), device='cpu', dtype=torch.float32)

        with torch.no_grad():
            out, pos_emb = relpos_enc(x)
            assert (out - x).sum().abs() <= 1e-8
            assert out.shape == x.shape

    @pytest.mark.unit
    @pytest.mark.parametrize('n_head', [1, 2, 10])
    @pytest.mark.parametrize('proj_dim', [None, -1])
    def test_transformer_mha_adapter_init(self, n_head, proj_dim):
        torch.random.manual_seed(0)
        x = torch.randn(2, 32, 50)
        lengths = torch.randint(1, x.size(1), size=(x.size(0),))
        lengths[torch.randint(0, x.size(0), size=(1,))[0]] = x.size(1)

        adapter = adapter_modules.TransformerMultiHeadAttentionAdapter(
            num_attention_heads=n_head, hidden_size=50, attn_layer_dropout=0.0, proj_dim=proj_dim
        )

        pad_mask, att_mask = get_mask(lengths)
        att_mask = att_mask.unsqueeze(1)

        with torch.no_grad():
            assert adapter.out_projection.weight.sum() == 0
            if hasattr(adapter.out_projection, 'bias') and adapter.out_projection.bias is not None:
                assert adapter.out_projection.bias.sum() == 0

            out = adapter(x, x, x, att_mask)
            assert out.sum().abs() <= 1e-8
            assert out.shape == x.shape

    @pytest.mark.unit
    def test_mha_adapter_strategy(self):
        adapter = adapter_modules.MultiHeadAttentionAdapter(n_head=1, n_feat=50, dropout_rate=0.0)
        assert hasattr(adapter, 'adapter_strategy')
        assert adapter.adapter_strategy is not None
        # assert default strategy is set
        assert isinstance(adapter.adapter_strategy, adapter_modules.MHAResidualAddAdapterStrategy)

    @pytest.mark.unit
    def test_relpos_mha_adapter_strategy(self):
        adapter = adapter_modules.RelPositionMultiHeadAttentionAdapter(n_head=1, n_feat=50, dropout_rate=0.0)
        assert hasattr(adapter, 'adapter_strategy')
        assert adapter.adapter_strategy is not None
        # assert default strategy is set
        assert isinstance(adapter.adapter_strategy, adapter_modules.MHAResidualAddAdapterStrategy)

    @pytest.mark.unit
    def test_abspos_encoding_adapter_strategy(self):
        adapter = adapter_modules.PositionalEncodingAdapter(d_model=50)
        assert hasattr(adapter, 'adapter_strategy')
        assert adapter.adapter_strategy is not None
        # assert default strategy is set
        assert isinstance(adapter.adapter_strategy, adapter_mixin_strategies.ReturnResultAdapterStrategy)

    @pytest.mark.unit
    def test_relpos_encoding_adapter_strategy(self):
        adapter = adapter_modules.RelPositionalEncodingAdapter(d_model=50)
        assert hasattr(adapter, 'adapter_strategy')
        assert adapter.adapter_strategy is not None
        # assert default strategy is set
        assert isinstance(adapter.adapter_strategy, adapter_mixin_strategies.ReturnResultAdapterStrategy)

    @pytest.mark.unit
    def test_transformer_mha_adapter_strategy(self):
        adapter = adapter_modules.TransformerMultiHeadAttentionAdapter(
            num_attention_heads=1, hidden_size=50, attn_layer_dropout=0.0
        )
        assert hasattr(adapter, 'adapter_strategy')
        assert adapter.adapter_strategy is not None
        # assert default strategy is set
        assert isinstance(adapter.adapter_strategy, adapter_modules.MHAResidualAddAdapterStrategy)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict

import torch

from nemo.collections.asr.models import EncDecSpeakerLabelModel
from nemo.collections.speechlm2.parts.precision import fp32_precision


class SECS:
    """
    Computes Speacker encoder cossine similarity (SECS) on generated audio with pretrained speaker encoder model.
    """

    def __init__(self, pretrained_se_name: str) -> None:
        self.speaker_encoder = None  # load into memory on reset()
        self.pretrained_se_name = pretrained_se_name
        self._secs = defaultdict(list)

    def reset(self):
        # Cleaning up GPU memory before we load ASRModel, because it may already
        # be quite fragmented and close to the limit after observing many
        # dynamic shapes during the training epoch.
        torch.cuda.memory.empty_cache()
        with fp32_precision():
            self.speaker_encoder = EncDecSpeakerLabelModel.from_pretrained(model_name=self.pretrained_se_name).eval()

        return self

    def update(
        self,
        name: str,
        target_audio: torch.Tensor,
        target_audio_lens: torch.Tensor,
        pred_audio: torch.Tensor,
        pred_audio_lens: torch.Tensor,
    ) -> None:
        if self.speaker_encoder is None:
            self.reset()

        with fp32_precision():
            with torch.no_grad():
                _, t_g = self.speaker_encoder(input_signal=target_audio, input_signal_length=target_audio_lens.long())
                _, s_g = self.speaker_encoder(input_signal=pred_audio, input_signal_length=pred_audio_lens.long())
            secs = torch.nn.functional.cosine_similarity(t_g, s_g, dim=-1).mean()

        self._secs[name].append(secs)

    def compute(self) -> dict[str, torch.Tensor]:
        """Computes the final score and deallocates ASR and partial results."""
        corpus_metric = {}
        avg_secs = []
        for name in self._secs.keys():
            secs = torch.stack(self._secs[name]).mean()
            corpus_metric[f"secs_{name}"] = secs
            avg_secs.append(secs)

        corpus_metric["secs"] = torch.stack(avg_secs).mean()
        self._secs.clear()
        self.speaker_encoder = None  # free up GPU memory
        torch.cuda.memory.empty_cache()
        return corpus_metric
